#include <opencv2/opencv.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/features2d/features2d.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/nonfree/nonfree.hpp>

#include <chrono>
#include <iostream>
#include <string>
#include <vector>
#include <fstream>
#include <iterator>
#include <map>
#include <algorithm>

typedef cv::Ptr<cv::FeatureDetector> FDetectPtr;
typedef cv::Ptr<cv::DescriptorMatcher> FMatchPtr;
typedef cv::Ptr<cv::DescriptorExtractor> FExtractPtr;

void printMat(const cv::Mat& m) {
    std::cout << m << std::endl;
}

void skipFrames(cv::VideoCapture c, int n) {
    for(int i = 0; i < n; i++) {
        if(!c.grab()) break;
    }
}

template <typename T>
void vectorToFile(std::string filename, std::vector<T> data) {
    std::ofstream outfile(filename.c_str());
    std::ostream_iterator<T> output_iterator(outfile, "\n");
    std::copy(data.begin(), data.end(), output_iterator);
}

double histogramDiff(cv::Mat im1, cv::Mat im2) {
    int histSize = 256;
    float range[] = {0, 256};
    const float* histRange = { range };
    bool uniform = true; bool accumulate = false;
    cv::Mat h1, h2; 
    cv::calcHist(&im1, 1, 0, cv::Mat(), h1, 1, &histSize, &histRange, uniform, accumulate);
    cv::calcHist(&im2, 1, 0, cv::Mat(), h2, 1, &histSize, &histRange, uniform, accumulate);
    return cv::compareHist(h1, h2, CV_COMP_HELLINGER);
}

cv::Point2f projectPoint(const cv::Mat& H, const cv::Point2f& p) {
    cv::Mat pHom(3,1,CV_64F);
    pHom.at<double>(0,0) = static_cast<double>(p.x);
    pHom.at<double>(1,0) = static_cast<double>(p.y);
    pHom.at<double>(2,0) = 1.0;
    cv::Mat pProj = H*pHom; 
    return cv::Point2f(pProj.at<double>(0,0)/pProj.at<double>(2,0),
            pProj.at<double>(1,0)/pProj.at<double>(2,0));
}

double computeReprojectionError(const cv::Mat& H, const std::vector<cv::Point2f>& srcPoints, 
        const std::vector<cv::Point2f>& dstPoints, const cv::Mat& inlierMask) {
    double err = 0.0;
    int numInliers = 0;
    for(int i = 0; i < srcPoints.size(); i++) {
        if(inlierMask.at<int>(i)==1) {
            cv::Point2f p1 = srcPoints[i];
            cv::Point2f p2 = dstPoints[i];
            cv::Point2f p1Proj = projectPoint(H, p1);
            err += cv::norm(p2-p1Proj);
            numInliers++;
        }
    }
    return err/numInliers;
}

int main(int argv, char **argc) {
    if(argv != 6) {
        std::cout << 
            "usage: ./featurematch <video file> <n frames to skip> <detector> <framerate> <max frames>" 
            << std::endl;
        return 1;
    }
    
    /* Open video */
    std::string vidFile(argc[1]);
    std::cout << "Opening video file: " << vidFile << std::endl;
    cv::VideoCapture vidCap(vidFile);
    
    /* Skip start frames */
    int nFrameSkip = std::atoi(argc[2]);
    skipFrames(vidCap, nFrameSkip);

    /* Create windows */
    std::string streamName("Video stream");
    std::string matchName("Feature matches");
    cv::namedWindow(streamName);
    cv::namedWindow(matchName);

    /* Create feature detector */
    std::string detectorStr(argc[3]);
    FDetectPtr detector = cv::FeatureDetector::create(detectorStr);
    
    /* Create feature extractor */
    FExtractPtr extractor = cv::DescriptorExtractor::create(detectorStr);

    /* Create feature matcher */
    std::map<std::string, std::string> detectMatchMap {
        {"ORB", "BruteForce-Hamming"},
        {"SIFT", "BruteForce"},
        {"SURF", "BruteForce"},
    };
    std::string matcherStr = detectMatchMap[detectorStr];
    FMatchPtr matcher = cv::DescriptorMatcher::create(matcherStr);
    
    /* Get FPS */
    double vidFps = vidCap.get( CV_CAP_PROP_FPS );
    double framerate = std::atof(argc[4]);
    int delay_ms = static_cast<int>(1E3/framerate);
    std::cout << "Video fps: " << vidFps << ", delay_ms: " << delay_ms << std::endl;

    /* Get max frames */
    int maxFrames = std::atoi(argc[5]);

    std::vector<double> imageDiffs;
    std::vector<double> histDiffs;
    std::vector<double> reprojectionErrors;
    std::vector<int> nKeypoints;
    std::vector<int> mKeypoints;
    std::vector<int> nMatches;
    std::vector<double> timing;
    cv::Mat currImage;
    cv::Mat currFeatures;
    cv::Mat correspondanceImage;
    std::vector<cv::KeyPoint> currKeypoints;
    int frameSkip = 5;
    int numKeypoints = 0;
    int nAcceptedFrames = 1;
    std::size_t maxKeypoints = 0;
    for(int frameNum = 0;frameNum < maxFrames;frameNum++) {
        /* Store previous image and keypoints */
        cv::Mat prevImage = currImage.clone();
        cv::Mat prevFeatures = currFeatures.clone();
        std::vector<cv::KeyPoint> prevKeypoints(currKeypoints);
        
        /* Get new image */
        cv::Mat rawImage;
        vidCap >> rawImage;
        cv::cvtColor(rawImage, currImage, CV_BGR2GRAY);

        /* Compute image difference, detect duplicates */
        if(frameNum > 0) {
            double imageDiff = cv::norm(currImage - prevImage);
            double histDiff = histogramDiff(currImage, prevImage);
            imageDiffs.push_back(imageDiff);
            histDiffs.push_back(histDiff);
            if(histDiff < 0.03){ // Duplicate! Skip if histogram diff is smaller than this
                continue;
            } else {
                nAcceptedFrames++;
            }
        }

        /* Start timing */
        auto start = std::chrono::system_clock::now();

        /* Detect keypoints */
        detector->detect(currImage, currKeypoints);

        /* Extract features */
        extractor->compute(currImage, currKeypoints, currFeatures);
        
        /* Match keypoints */
        if(frameNum > 0 && currKeypoints.size() > 0 && prevKeypoints.size() > 0) {
            std::vector<cv::DMatch> matches;
            matcher->match(currFeatures, prevFeatures, matches);
            /* Filter out "bad" matches */
            std::sort(matches.begin(), matches.end());
            cv::DMatch minElem = matches.front();
            matches.erase(
                    std::remove_if(matches.begin(), 
                        matches.end(),
                        [=](cv::DMatch d){return d.distance > 2*minElem.distance;}),
                    matches.end());

            cv::Mat inlierMask = cv::Mat();
            /* Estimate homography */
            if(matches.size() >= 4) {
                nMatches.push_back(matches.size());
                std::vector<cv::Point2f> currPoints;
                std::vector<cv::Point2f> prevPoints;
                for(auto m : matches) {
                    currPoints.push_back(currKeypoints[m.queryIdx].pt);
                    prevPoints.push_back(prevKeypoints[m.queryIdx].pt);
                }
                cv::Mat H = cv::findHomography(currPoints, prevPoints, 
                        CV_RANSAC, 3, inlierMask);
                /* Compute reprojection error */
                double reprojError = computeReprojectionError(H, currPoints, prevPoints, inlierMask);
                reprojectionErrors.push_back(reprojError);
                //std::cout << "Reprojection error: " << reprojError  
                 //   << ", num inliers: " << cv::sum(inlierMask) << std::endl;
            }
            cv::drawMatches(currImage, currKeypoints,
                            prevImage, prevKeypoints,
                            matches, correspondanceImage,
                            cv::Scalar::all(-1), cv::Scalar::all(-1), inlierMask);
            imshow(matchName, correspondanceImage);
        }
        /* End timing */
        auto end = std::chrono::system_clock::now();
        std::chrono::duration<double> elapsed_seconds = end-start;
        timing.push_back(elapsed_seconds.count());

        
        /* Draw keypoints */
        cv::Mat drawImage = currImage.clone();
        cv::drawKeypoints(drawImage, currKeypoints, drawImage);

        /* Compute keypoint metrics */
        numKeypoints += currKeypoints.size();
        maxKeypoints = std::max(maxKeypoints, currKeypoints.size());
        if(nAcceptedFrames % frameSkip == 0) {
            std::cout << "Frame number: " << frameNum 
                << ", keypoints/frame: " << numKeypoints/frameSkip 
                << ", max keypoints: " << maxKeypoints << std::endl;
            nKeypoints.push_back(numKeypoints/frameSkip);
            mKeypoints.push_back(maxKeypoints);
            numKeypoints = 0;
            maxKeypoints = 0;
        }

        /* Show stream */
        imshow(streamName, drawImage);
        cv::waitKey(0*delay_ms);
        prevImage = currImage;
    }

    /* Store data */
    std::string filePrefix = "data/" + detectorStr + "-" + 
        "FS" + std::to_string(frameSkip) + "-" +
        "NF" + std::to_string(nAcceptedFrames);
    vectorToFile(filePrefix+"-nKeypoints.csv", nKeypoints);
    vectorToFile(filePrefix+"-mKeypoints.csv", mKeypoints);
    vectorToFile(filePrefix+"-nMatches.csv", nMatches);
    vectorToFile(filePrefix+"-reprojectionErrors.csv", reprojectionErrors);
    vectorToFile(filePrefix+"-imageDiff.csv", imageDiffs);
    vectorToFile(filePrefix+"-imageHist.csv", histDiffs);
    vectorToFile(filePrefix+"-timing.csv", timing);

    return 0;
}
