#ifndef MAVLINK_INTERFACE_H
#define MAVLINK_INTERFACE_H

#include <ros/ros.h>
#include <std_msgs/String.h>
#include <tf/transform_listener.h>
#include <tf/transform_datatypes.h>

#include <cmath>
#include <string>
#include <sstream>
#include <chrono>

#define MAVLINK_DIALECT ardupilotmega
#include <mavconn/interface.h>

namespace Snakelab {

class MavlinkInterfaceNode
{
	public:
		MavlinkInterfaceNode();
		void run();
	private:
		ros::NodeHandle n;
		ros::Publisher cmdPub;
		std::string url;
		std::chrono::milliseconds boot_time;
		mavconn::MAVConnInterface::Ptr mavInterfacePtr;
		tf::TransformListener poseListener;
		std::string world_frame_ned;
		std::string world_frame_ecef;
		std::string body_frame;
		std::string world_frame_mocap;
		void recv_message(const mavlink_message_t *msg, uint8_t sysid, uint8_t compid);
		void send_heartbeat();
		void send_nav_controller_packet();
		void send_waypoint_ack();
		void send_local_pose(const tf::StampedTransform& transform, uint32_t time_msec);
		void send_global_position(const tf::StampedTransform& transform, uint32_t time_msec);
};

/* WGS84 ellipsoidal parameters */
#define WGS84_RE 6378137.0 // major axis
#define WGS84_RP 6356752.0 // minor axis

/* Compute latitude, longitude and height above WGS84 ellipsoid from ECEF coordinates iteratively,
 * implementation from Algorithm 2.4 in Fossen 2011.
 * Input: ECEF x, y, z in meters
 * Output: latitude and longitude in degrees, height above ellipsoid in meters 
 * Andreas L. Flåten, 2015 */
void ECEF_to_llh(const double &x, const double &y, const double &z, double &lat_deg, double &lon_deg, double &hae);

/* Compute ECEF coordinates from latitude, longitude and height above WGS84 ellipsoid
 * implementation from Fossen 2011.
 * Input: latitude and longitude in degrees, height above ellipsoid in meters 
 * Output: ECEF x, y, z in meters
 * Andreas L. Flåten, 2015 */
void llh_to_ECEF(const double &lat_deg, const double &lon_deg, const double &hae, double &x, double &y, double &z);

} // namespace Snakelab

#endif
