#ifndef GRIDMAPPING_H
#define GRIDMAPPING_H

#include <string>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/core/core.hpp>

namespace AMOS {

class CameraModel {
	public:
		CameraModel(const std::string& paramFilename); // the expected format of the file is an openCV camera calibration XML/YAML file
		const cv::Matx33f getCamMatrix() { return camMatrix_; };
		const cv::Mat getDistortionCoeffs() { return distortionCoeff_; };
		void undistortImage(const cv::Mat& input, cv::Mat& output);
	private:
		cv::Matx33f camMatrix_;
		cv::Mat distortionCoeff_;
};

/* The grid map is a planar occupancy grid map with a finite number of grid cells. The occupancy probabilities are represented in log-odds form.
 * I.e. map_(i,j) = log(P(grid cell i,j occupied)/P(grid cell i,j not occupied)). 
 * A value of zero means equal probability, while a value of infinity implies that P(grid cell i,j occupied) = 1 */
class GridMap {
	private:
		double dx_; // metric grid size in the x direction (meters)
		double dy_; 
		uint32_t width_; // Map grid width (num cells in the x direction)
		uint32_t height_; 
		cv::Point3f originWorldMap_; // map origin relative to a world fixed frame, expressed in the world frame
		cv::Matx33f dcmWorldMap_; // map rotation matrix from a world fixed frame to the map frame
		cv::Matx33f mapMatrix_; // transformation from planar homogeneous map coordinates (u',v',1) metric map coordinates (x_m,y_m,z_m)
		cv::Mat map_; // grid map representation in log odds form
	public:
		static const int MAP_DTYPE = CV_32F;
		GridMap() : 
			GridMap(250, 250) {};
		GridMap(const uint32_t width, const uint32_t height) : 
			GridMap(width, height, 0.025, 0.025) {};
		GridMap(const uint32_t width, const uint32_t height, const double dx, const double dy) :
			GridMap(width, height, dx, dy, cv::Point3f(-1.5f,0.5f,0), cv::Matx33f(cv::Mat(cv::Mat::eye(3,3,CV_32F))) ) {};
		GridMap(const uint32_t width, const uint32_t height, const double dx, const double dy, 
			const cv::Point3f origin, const cv::Matx33f dcmWorldMap) : 
			dx_(dx), dy_(dy), width_(width), height_(height),
			originWorldMap_(origin), dcmWorldMap_(dcmWorldMap),
			mapMatrix_(dx, 0, 0.5*dx,
				   0, dy, 0.5*dy,
				   0, 0, 0),
			map_(cv::Mat::zeros(width, height, MAP_DTYPE)) {};
		uint32_t getWidth() { return width_; };
		uint32_t getHeight() { return height_; };
		cv::Size getSize() { return cv::Size(width_, height_); };
		cv::Point2i toMapCoordinates(cv::Point2f p_map) { return cv::Point2i(int(p_map.x/dx_), int(p_map.y/dy_)); }
		cv::Point3f toMapFrame(cv::Point3f pointWorld);
		void addMeasurement(const cv::Mat measurement);
		/* Returns a matrix that transform a map homogeneous coordinate 
		 * (u',v',1) to a world fixed coordinate (x_w, y_w, z_w) */
		const cv::Matx33f getMapWorldTransform();
		const cv::Mat& getMap() { return map_; };
};

// Simple struct to represent ground-truth ice features in the world frame
struct IceFeature {
	IceFeature(float x, float y, float r) : r(r), posWorld(x, y, 0.0) {};
	cv::Point3f posWorld;
	float r;
};

class RecursiveMapEstimator {
	public:
		RecursiveMapEstimator(const std::string& paramFilename) : 
			gridMap_(300, 300, 0.010, 0.010), camModel_(paramFilename) {};
		// Update map estimate based on a segmented image. Expects a binary segmented image and camera pose as input
		void updateMap(cv::Mat& segmentedImage, const cv::Matx33f& dcmCamWorld, const cv::Point3f& originCamWorld);
		// Extra functions for testing
		void projectImageToMap(const cv::Mat& input, cv::Mat& output, const cv::Matx33f dcmCamWorld, const cv::Point3f originCamWorld);
		void getProbMap(cv::Mat& output);
		void drawFeaturesOnMap(std::vector<IceFeature> features, cv::Mat& output);
	private:
		GridMap gridMap_;
		CameraModel camModel_;
		const cv::Matx33f getPerspectiveTransform(const cv::Matx33f dcmCamWorld, const cv::Point3f originCamWorld);
		void toLogOdds(cv::Mat& inputProb, cv::Mat& outputLogOdds);
		void toProbability(cv::Mat& inputLogOdds, cv::Mat& outputProbability);
		void segmentedToProb(cv::Mat& segmentedImage, cv::Mat& detectionProb);
};

} // namespace AMOS

#endif
